from itertools import combinations, chain
from sliding_window import window

def _factorial(f, iterable, n_columns):
    repeatable = list(iterable)
    for i in range(1, 1 + n_columns):
        yield from f(repeatable, i)

def _fromdicts(header, data, n_columns, only_adjacent):
    '''
    Given an iterable of dicts, find the keys that are primary keys across the dicts.

    Args:
        data: Iterable of dicts
        n_columns: How many columns wide is the primary key allowed to be?
        only_adjacent: If this is true, I'll assume that columns in a primary key must be adjacent.

    Returns:
        set of primary keys
    '''
    return _dedupe(_fromdicts_duplicates(header, data, n_columns, only_adjacent))

def _fromdicts_duplicates(header, data, n_columns, only_adjacent):
    if len(header) < n_columns:
        n_columns = len(header)

    if only_adjacent:
        f = window
    else:
        f = combinations

    hashes = {columns:set() for columns in _factorial(f, header, n_columns)}
    nrow = 0
    for row in data:
        for columns in hashes:
            hashes[columns].add(multicol_hash(row, columns))
        nrow += 1
    return set(k for k,v in hashes.items() if len(v) == nrow)

def _dedupe(indices:set) -> set:
    '''
    Remove complex indices when simpler equivalents exist.
    '''
    result = set()
    top = 1 if indices == set() else max(map(len, indices)) 
    for ncol in range(1, top + 1):
        for columns in indices:
            if len(columns) == ncol:
                for subcolumns in _factorial(combinations, columns, ncol):
                    if subcolumns in result:
                        break
                else:
                    result.add(columns)
    return result

def multicol_hash(row, columns):
    return hash(tuple(row[column] for column in columns if column != None))



import os
import io
import csv

from special_snowflake.fromcsv import _separate_fp, _dialect
from special_snowflake.fromdicts import _fromdicts

def fromcsv(fp, n_columns = 3, only_adjacent = False, **csv_kwargs):
    # Check for emptiness
    position = fp.tell()
    if fp.seek(0, os.SEEK_END) == 0:
        return set()
    fp.seek(position, os.SEEK_SET)

    # Detect dialect if no csv stuff is specified
    if csv_kwargs == {}:
        try:
            csv_kwargs['dialect'] = _dialect(fp)
        except csv.Error:
            pass

    # Go
    header, data = _separate_fp(fp, **csv_kwargs)
    return _fromdicts(header, data, n_columns, only_adjacent)

def fromresponse(response, n_columns = 3, only_adjacent = False, **csv_kwargs):
    'Load from an HTTP response object.'
    fp = io.StringIO(response.text)
    return fromcsv(fp, n_columns = n_columns, only_adjacent = only_adjacent, **csv_kwargs)

def fromdicts(header, data, n_columns = 3, only_adjacent = False):
    return _fromdicts(header, data, n_columns, only_adjacent)



import csv
import json

def _separate_fp(fp, *args, **kwargs):
    '''
    Return a tuple of
        list of field names
        csv.DictReader
    '''
    position = fp.tell()
    header = next(csv.reader(fp, *args, **kwargs))
    fp.seek(position)

    _kwargs = dict(kwargs)
    _kwargs['fieldnames'] = _kwargs.get('fieldnames', header)
    _kwargs['restval'] = _kwargs.get('restval', '')
    data = csv.DictReader(fp, *args, **_kwargs)
    next(data) # skip the header
    return header, data

def _dialect(fp):
    'Guess the dialect of a CSV file.'
    pos = fp.tell()
    dialect = csv.Sniffer().sniff(fp.read(1024))
    fp.seek(pos)
    return dialect



#!/usr/bin/env python3
from distutils.core import setup

setup(name='special_snowflake',
    author = 'Thomas Levine',
    author_email = '_@thomaslevine.com',
    description = 'Find the unique indices for a dataset',
    url = 'https://github.com/tlevine/special_snowflake.git',
    classifiers = [
        'Intended Audience :: Developers',
    ],
    packages = ['special_snowflake'],
    install_requires = ['sliding-window'],
    tests_require = ['nose'],
    version = '0.0.9',
    license = 'AGPL',
)
